/*
 * iwconnect/iwc_notify.c
 * $Rev: 1222 $
 *
 * Copyright (C) 2009-2010 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/**
 * @file iwc_notify.c
 * @brief Functions for handling creation, polling and removal of a IPC message
 * queue for receiving vif parameter update notifications
 */

/* INCLUDES *******************************************************************/

#include <netinet/in.h>                 // struct in_addr
#include <linux/if.h>                   // IFNAMSIZ
#include <sys/ipc.h>                    // IPC_CREATE
#include <sys/msg.h>                    // msgget(), msgrecv(), msgsnd(), ...
#include <errno.h>                      // errno
#include <stdio.h>                      // printf()
#include <string.h>                     // memset(), strerror()
#include "iwc_print.h"


/* MESSAGE QUEUE TYPES ********************************************************/

#define VIF_NOTIFY_KEY      1337    /**< Name of notification message queue */

/** Message format to indicate a ioctl() */
typedef struct {
    long mtype;                     /**< Message type */
    int ioctl_type;                 /**< ioctl() type */
    char dev_name[IFNAMSIZ + 1];    /**< Virtual interface device name */
} ioctl_msg;

#define IOCTL_MSG_SIZE      sizeof(ioctl_msg)
#define VIF_SIOCSIW         0x10    /**< Message type: ioctl() set wireless parameter */


/* FUNCTIONS ******************************************************************/

/**
 * Initialize notification message queue
 * @return Notification queue ID
 */
int
iwc_notify_init()
{
    int create_flags = IPC_CREAT | 0664;

    int old_id = msgget(VIF_NOTIFY_KEY, create_flags | IPC_EXCL);
    if (errno == EEXIST)                         // check for previous msg queue
    {
        old_id = msgget(VIF_NOTIFY_KEY, 0664);
        msgctl(old_id, IPC_RMID, NULL);         // delete old message queue
    }
    return msgget(VIF_NOTIFY_KEY, create_flags);
}

/**
 * Delete notification message queue
 * @param msgq_id Message queue ID
 */
int
iwc_notify_remove(int msgq_id)
{
    return msgctl(msgq_id, IPC_RMID, NULL);
}

/**
 * Poll the notification queue for arriving messages. On message reception run
 * the given msg_handler function with the message content
 * @param msg_handler Pointer to a client function handling parameter changes.
 * Its signature is void func(int ioctl_type, char *dev_name)
 */
void
iwc_notify_poll(void *msg_handler)
{
    ioctl_msg in_buffer;
    int msgq_id = -1;

    if (msg_handler == NULL)
    {
        iwc_error("iwc_notify_init(): msg_handler == NULL\n");
        return;
    }

    // initialize message queue where we will listen for updates
    iwc_notify_init();
    memset(&in_buffer, 0, sizeof(ioctl_msg));
    while (1)
    {
        // inside loop, because value is overwritten at msgrcv(?!)
        msgq_id = msgget(VIF_NOTIFY_KEY, 0400);
        if (msgrcv(msgq_id, &in_buffer, IOCTL_MSG_SIZE, VIF_SIOCSIW, 0) < 0)
        {
            iwc_error("iwc_notify_poll(): msgrcv (%s)\n", strerror(errno));
            return;
        }
        // now correctly cast the the function pointer and execute
        ((void (*)(int ioctl_type, const char *dev_name)) msg_handler)
            (in_buffer.ioctl_type, in_buffer.dev_name);
    }
    return; // this should never be reached!
}